#define TITLE "z - Nigel's fast viewer - nib 1998-08-02"
/*
       z - a fast pager for ascii files on IBM-compatible PCs
       ------------------------------------------------------

z is intended to allow rapid examination of large or small ascii files
on fully IBM-compatible PCs.  The command set is based on the Unix
utility 'less', though the code is completely different.  Extra features
have been added to support PC-specific functions, and to make the pager
more convenient for reading large ascii texts such as whole books, where
the file may be read in more than one session.

Fast reader for big files - nib - started 1994-01-22

Must use model with far data pointers (COMPACT or LARGE).
*/
#define COPR  "Copyright (c) 1994,1998 Nigel Bromley - +44 7010 700642"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <alloc.h>
#include <stdarg.h>
#include <io.h>
#include <dir.h>
#include <dos.h>
#include <fcntl.h>
#include <setjmp.h>
#include "regex.h"

/* globals */

#define WANTED_MODE 0x03            	/* use 24 * 80 colour text screen */
#define NORM_MEM ((char far *)0xb8000000)
#define HERC_MEM ((char far *)0xb0000000)
#define COLUMNS 80
#define DEFAULTROWS 25
#define BOOKMARKS 32 			/* don't change without thought ! */
#define DEFAULTUDMOVE 10
#define BUFFERSIZE (COLUMNS*400)
#define MAXBUFFERS 40
#define GLOBALTABLESIZE 4096		/* careful, this*10 < 64000k */
#define DISKBUFFER 32768

#define FF 0x0c
#define BS 0x08
#define READ_KEYB_SUBFUNC 0x00
#define READ_KEYB_INT 0x16

#define BIGGEST (0x7ffffff0L)
#define HOME 0x4700
#define END 0x4f00
#define PAGEUP 0x4900
#define PAGEDOWN 0x5100
#define CURRDOWN 0x5000
#define CURRUP 0x4800
#define ESCAPE 0x001b
#define F1 0x3b00
#define CONTROL_C 0x03

int in_file;
int uselines, statusline, rows;
char far *disp_mem;
long udmove;
union REGS regs;
char mine[] = {"(c) 1994,1998 Nigel Bromley, +44 1908 261740"};
long last_line;
int buffers;
int current_buff;
long mark[BOOKMARKS];
int bios;
int attrib;
int ignore_case;
int save_mode;
long serial;
unsigned char search_key[COLUMNS + 2] = {""};
unsigned char *save_screen, *disk_buffer;
jmp_buf jump_buff;

/* --------------------------------------------------------------------------
Text location structures.  The global table has one entry for each roughly
BUFFERSIZE segment into which the file is divided.  Each of these may or may
not be in core at any time.  Each block starts at a line start, the line-
number of this entry appears as 'first_line'.  'Offset' is the location
of this line start in the file.  Entries are in numerical order of line
number.

The local table has one entry for each core buffer, giving a pointer to its
first location.  Access-serial is used for a least-recently used allocation
system.  Segment is a back-link to the global_table.
*/

struct global_table_entry {
	long first_line;
	long file_offset;
	int in_buffer;
} *global_table;

struct local_table_entry {
	char *buffer_location;
	int segment;
	long access_serial;
} *local_table;

/* --------------------------------------------------------------------------
Wildcard expansion for DOS, pinched from PD vi-like editor "elvis" by:
 *	Guntram Blohm
 *	Buchenstrasse 19
 *	7904 Erbach, West Germany
 *	Tel. ++49-7305-6997
Modified by nib - 1993-03-12
Takes a pointer to argc together with argv, modifies argc and returns a new
char ** which should be used in place of argv. It mallocs itself enough space,
which cannot be released.
*/

#define	MAXFILES 1000

char *files[MAXFILES];
int nfiles;

int pstrcmp(const void *a, const void *b)
{
	return strcmp(*(char **)a, *(char **)b);
}

void addfile(char *buf)
{
	char *p;

	for (p=buf; *p; p++)
		*p = tolower(*p);

	if (nfiles<MAXFILES && (files[nfiles]=malloc(strlen(buf)+1))!=0)
		strcpy(files[nfiles++], buf);
}

void expand(char *name)
{
	char *filespec;
	int wildcard = 0;
	struct ffblk findbuf;
	int err;
	char buf[80];
	int lastn;

	strcpy(buf, name);
	for (filespec=buf; *filespec; filespec++)
		;

	while (--filespec >= buf)
	{	if (*filespec == '?' || *filespec == '*')
			wildcard=1;
		if (*filespec == '/' || *filespec == '\\' || *filespec == ':')
			break;
	}
	if (!wildcard)
		addfile(buf);
	else
	{
		lastn = nfiles;
		filespec++;
		if ((err=findfirst(buf, &findbuf, 0)) != 0)
			addfile(buf);
		while (!err)
		{
			strcpy(filespec, findbuf.ff_name);
			addfile(buf);
			err = findnext(&findbuf);
		}
		if (lastn != nfiles)
			qsort(files+lastn, nfiles-lastn, sizeof(char *),
								pstrcmp);
	}
}

char **wildexpand(int *pargc, char **argv)
{
	int i;
	
	for (i=0; i<*pargc; i++)
		expand(argv[i]);
	*pargc = nfiles;
	return files;
}	

/* --------------------------------------------------------------------------
BIOS call to set video mode.
*/
void set_video_mode(int mode)
{
#define VIDEO_INT 0x10
#define SET_VIDEO_MODE_SUBFUNC 0x00

	union REGS regs;

	regs.h.ah = SET_VIDEO_MODE_SUBFUNC;
	regs.h.al = mode;
	int86(VIDEO_INT, &regs, &regs);
}

/* --------------------------------------------------------------------------
BIOS call to detect video mode.
*/
int get_video_mode(void)
{
#define GET_CURR_DIS_MODE_SUBFUNC 0x0f

	union REGS regs;

	regs.h.ah = GET_CURR_DIS_MODE_SUBFUNC;
	int86(VIDEO_INT, &regs, &regs);
	return regs.h.al;
}

/* --------------------------------------------------------------------------
Get a key directly from the keyboard. This should not be necessary, it
isn't in Unix, but I can't at the moment find any other way of getting
unbuffered character input from the keyboard while stdin is redirected
elsewhere. Getch() works brilliantly normally, but, despite the blurb,
it seems to stay with stdin when that is redirected, and not, as stated,
with the keyboard.
*/
unsigned get_keyb(void)
{
	regs.h.ah = READ_KEYB_SUBFUNC;
	int86(READ_KEYB_INT, &regs, &regs);
	return regs.x.ax;
}

/* --------------------------------------------------------------------------
BIOS call to check if there is a key waiting, returns non-zero if there is.
*/
unsigned int check_keyb(void)
{
#define READ_KEYB_STAT_SUBFUNC 0x01

	union REGS regs;

	regs.h.ah = READ_KEYB_STAT_SUBFUNC;
	int86(READ_KEYB_INT, &regs, &regs);
	if (regs.x.flags & 0x0040)   /* z-flag set */
		return 0;
	else
		return -1;
}

/* --------------------------------------------------------------------------
Write character at x, y
*/
void write_char(int x, int y, int ch)
{
	if (bios)
	{
                regs.h.ah = 0x02;	/* set cursor posn */
		regs.h.bh = 0;
		regs.h.dh = y;
		regs.h.dl = x;
		int86(VIDEO_INT, &regs, &regs);
		regs.h.ah = 0x0a;	/* write char at cursor */
		regs.h.al = ch;
		regs.h.bh = 0;
		regs.h.bl = 0;
		regs.x.cx = 1;
		int86(VIDEO_INT, &regs, &regs);
	}
	else
	{
		*(disp_mem + 160*y + 2*x) = ch;
	}
}

/* --------------------------------------------------------------------------
Write line at y. Str must be long enough!
*/
void write_line(int y, char *str)
{
	int i;
	char far *p;

	if (bios)
	{
		for (i=0; i<COLUMNS; i++)
			write_char(i, y, *(str+i));
	}
	else
	{
		i = COLUMNS;
              	p = disp_mem + 160*y;
		while (i--)
		{
			*p++ = *str++;
			p++;
		}
	}
}

/* --------------------------------------------------------------------------
Clear line y
*/
void clear_line(int y)
{
	int i;
	char far *p;

	if (bios)
	{
                regs.h.ah = 0x02;	/* set cursor posn */
		regs.h.bh = 0;
		regs.h.dh = y;
		regs.h.dl = 0;
		int86(VIDEO_INT, &regs, &regs);
		regs.h.ah = 0x0a;	/* write char at cursor */
		regs.h.al = ' ';
		regs.h.bh = 0;
		regs.h.bl = 0;
		regs.x.cx = COLUMNS;
		int86(VIDEO_INT, &regs, &regs);
	}
	else
	{
		i = COLUMNS;
              	p = disp_mem + 160*y;
		while (i--)
		{
			*p++ = ' ';
			*p++ = (y == statusline) ?
				(attrib & 0x88) | ((attrib >> 4) & 0x07)
				| ((attrib << 4) & 0x70) : attrib;
		}
	}
}

/* --------------------------------------------------------------------------
Clear screen
*/
void clear_screen(void)
{
	int i;

	for (i=0; i<uselines; i++)
		clear_line(i);
	clear_line(statusline);
}

/* --------------------------------------------------------------------------
Put a formatted string at given x, y, Returns the length of the string.
Uses vsprintf to convert the string, then write_char to write it.
*/
int dprintf(int x, int y, char *format, ...)
{
	char s[100], *t;
	va_list args;

	va_start(args, format);
	vsprintf(s, format, args);
	clear_line(y);
	t = s;
	while (*t)
		write_char(x++, y, *t++);
	va_end(args);

	return strlen(s);
}

/* --------------------------------------------------------------------------
Write 'uselines' lines each of a dump of 16 bytes, starting at 'address'.
Stop early if would be greater than file_size
*/
void hex_screen(int filehandle, long address, long file_size)
{
	int i, j, lim;
	long x;
	unsigned char *p, *m, s[100];

	if (-1 == lseek(filehandle, address, SEEK_SET))
	{
		clear_screen();
		dprintf(1, 2, "seek error");
		return;
	}
	x = min(uselines * 16, file_size - address);
	i = _read(filehandle, disk_buffer, x);
	clear_screen();
	if (x != i)
	{
		dprintf(1, 2, "read error");
		return;
	}
	p = disk_buffer;
        for (i=0; i<uselines; i++)
        {
		m = s;
		if (0 == (lim = min(file_size - address, 16)))
			break;
		m += sprintf(m, " %08lx  ", address);
		for (j=0; j<lim; j++)
			m += sprintf(m, "%02x ", p[j]);
		for (; j<16; j++)
			m += sprintf(m, "   ");
		m += sprintf(m, " ");
		for (j=0; j<lim; j++)
			m += sprintf(m, "%c", p[j]);
		for (; j<16; j++)
			m += sprintf(m, " ");
		m += sprintf(m, "    ");
		write_line(i, s);
		p += lim;
		address += lim;
	}
}

/* --------------------------------------------------------------------------
Help for hex mode
*/
void help3(void)
{
	clear_screen();
	dprintf(1, 1, TITLE);
	dprintf(1, 2,  "  hex mode commands:");
	dprintf(1, 3,  "CR CurDn      one more line");
	dprintf(1, 4,  "SP PagDn      one more screen");
       	dprintf(1, 5,  "<n>g          goto a hex address");
	dprintf(1, 6,  "h F1          this help");
	dprintf(1, 7,  "q ESC         back to text mode");
	dprintf(1, 8,  "t HOME        top");
	dprintf(1, 9,  "z END         bottom");
	dprintf(1, 10, "CurUp y       back a line");
	dprintf(1, 11, "PagUp u       back a screen");
	dprintf(1, statusline, "CR to continue");
	get_keyb();
}

/* --------------------------------------------------------------------------
Do hex mode - a completely separate system
*/
void hex_mode(int in_file, long first_address, char *filename)
{
	int end;
	unsigned tc;
	long f_size, address, number;

	address = first_address;
	number = 0;
	end = 0;

	f_size = filelength(in_file);

	while (!end)
	{
		if (address < 0)
			address = 0;
		if (address >= f_size)
			address = f_size - 1;
		if (number > 0x7ffffffL)
			number = 0x7ffffffL;

		if (number == 0)
			hex_screen(in_file, address, f_size);

		dprintf(1, statusline,
			"z (hex mode) - %s - %ld bytes (0x%lx) [%lx]",
			filename, f_size, f_size, number);

		tc = get_keyb();
		if (tc & 0xff)
			tc &= 0xff;

		switch (tc)
		{
		case '0': case '1': case '2': case '3': case '4':
		case '5': case '6': case '7': case '8': case '9':
			number = number * 16 + (tc - '0');
			break;
		case 'a': case 'b': case 'c': case 'd':
		case 'e': case 'f':
			number = number * 16 + (toupper(tc)-'A'+10);
			break;
		case ' ':
		case PAGEDOWN:
			address += uselines * 16;
			number = 0;
			break;
		case 't':
		case HOME:
			address = 0;
			number = 0;
			break;
		case 'u':
		case PAGEUP:
			address -= uselines * 16;
			number = 0;
			break;
		case '\r':
		case CURRDOWN:
			address += 16;
			number = 0;
			break;
		case 'z':
		case END:
			address = ((f_size-1) & 0xfffffff0) -
				(uselines-1) * 16;
			number = 0;
			break;
		case 'g':
			address = number;
			number = 0;
			break;
		case CURRUP:
		case 'y':
			address -= 16;
			number = 0;
			break;
		case 'h':
		case F1:
			help3();
			number = 0;
			break;
		case 'q':
		case ESCAPE:
			end = 1;
			break;
		default:
			number = 0;
			break;
		}
	}
}

/* --------------------------------------------------------------------------
Fetch the next character from the input file.  Return -1 if no more.
If called with file == -1, resets the buffering.
*/
int next_char(int filehandle)
{
	static avail=0;
        static unsigned char *p;

        if (filehandle == -1)
	{
		avail = 0;
		return -1;
	}
	if (avail <= 0)
	{
		avail = _read(filehandle, disk_buffer, DISKBUFFER);
		p = disk_buffer;
		if (avail == 0)
			return -1;
	}
        avail--;
	return *p++;
}

/* --------------------------------------------------------------------------
Return a pointer to the line given, offset within the given buffer.
Line is assumed to be in the buffer.
*/
char *line_in_buffer(long line, int buffer)
{
	struct local_table_entry *p;

	p = &(local_table[buffer]);
	p->access_serial = ++serial;

	return p->buffer_location + line * COLUMNS;
}

/* --------------------------------------------------------------------------
Load a segment into a buffer - the one least recently used.  Given line
number of first line and file offset to read.  Returns by variable
the number of the next line (start of next segment), its offset, and the
number of the buffer used.  Return value 0 normally, or -1 if EOF found,
in which case nline is the number of the last line, and offset is undefined.

Note that line_in_file() also parses the file into lines, keep the two in
synch!  (I'll do it better next time.)
*/
int load_segment(long first_line, long offset, long *nline, long *noffset,
			int *buffer)
{
	long line, end, max, earliest;
	int ch, col;
	int i, victim, eof;
	char *p;

	while (check_keyb())
		if ((get_keyb() & 0xff) == CONTROL_C)
			longjmp(jump_buff, 1);

	lseek(in_file, offset, SEEK_SET);
        next_char(-1);
	end = offset;
	eof = 0;
	earliest = BIGGEST; victim = 0;
	max = BUFFERSIZE / COLUMNS;
	for (i=0; i<buffers; i++)
	{
		if (local_table[i].access_serial < earliest)
		{
			earliest = local_table[i].access_serial;
                	victim = i;
		}
	}
	if (local_table[victim].segment != -1)
		global_table[local_table[victim].segment].in_buffer = -1;
	local_table[victim].segment = -1;
	p = local_table[victim].buffer_location;
	local_table[victim].access_serial = ++serial;
	*buffer = victim;
	dprintf(1, statusline, "loading buffer %d starting line %ld",
			victim, first_line);
        ch = -1;
	for (line = first_line; line < first_line + max; line++)
	{
		col = 0;
		for (;;)
		{
			if (ch == -1)
			{
				ch = next_char(in_file);
			}
			switch (ch)
			{
			case -1:
                		while (col < COLUMNS)
				{
					*p++ = ' ';
					col++;
				}
                       		eof = 1;
				goto way_out;
			case '\r':
				ch = -1;
				end++;
				break;
			case '\t':
				*p++ = ' ';
				col++;
				while(col % 8 && col < COLUMNS)
				{
					*p++ = ' ';
					col++;
				}
				ch = -1;
				end++;
				if (col == COLUMNS)
					goto new_line;
				break;
			case '\n':
			case FF:
				while (col < COLUMNS)
				{
					*p++ = ' ';
					col++;
				}
				ch = -1;
				end++;
				goto new_line;
			default:
				if (col >= COLUMNS)
					goto new_line;
                                *p++ = ch;
                        	col++;
				ch = -1;
				end++;
				break;
			}
		}
	new_line:  ;
	}
way_out:
	*noffset = end;
	*nline = line;
	return eof ? -1 : 0;
}

/* --------------------------------------------------------------------------
Get a pointer to a line.  If error, returns a pointer to an error string
instead.  A line is of length COLUMNS (except the error string!), and
nulls are valid characters.
*/
char *point_line(long n)
{
	struct global_table_entry *p;
	int segment, next_segment, buffer;
	long foo, max_line, nline, noffset;

try_again:
	if (n > last_line)
		n = last_line;
	if (n < 0)
		n = 0;
	max_line = 0;
	next_segment = 0;
	for (segment=0; segment<GLOBALTABLESIZE; segment++)
	{
		p = &(global_table[segment]);
		if (p->first_line > n)
		{
			/* we've read this line before */
			p--;
			if (p->in_buffer == -1)
			{
				/* re-read the buffer */
				load_segment(p->first_line, p->file_offset,
					&foo, &foo, &buffer);
                 		p->in_buffer = buffer;
				local_table[buffer].segment = segment-1;
			}
			current_buff = p->in_buffer;
			return line_in_buffer(n - p->first_line,
							p->in_buffer);
		}
		if (p->first_line > max_line)
		{
			max_line = p->first_line;
			next_segment = segment;
		}
	}
	/* not in table yet - read another buffer and try again */
	if (next_segment > GLOBALTABLESIZE-2)
		return " File is too big!!   ";
        foo = load_segment(global_table[next_segment].first_line,
				global_table[next_segment].file_offset,
				&nline, &noffset, &buffer);
	local_table[buffer].segment = next_segment;
	global_table[next_segment+1].first_line = nline;
	global_table[next_segment+1].file_offset = noffset;
	global_table[next_segment].in_buffer = buffer;
	if (foo != 0)
	{
		last_line = nline;
		global_table[next_segment+1].first_line = last_line+1;
	}
	goto try_again;
}

/* --------------------------------------------------------------------------
Given a line_number, find a pointer to it within the file.  If for any
reason cannot find it, return zero (start of file)

Note that load_segment() also parses the file into lines, keep the two in
synch!  (I'll do it better next time.)
*/
long line_in_file(long line)
{
	long temp, file_p;
	unsigned char *p;
	int ch, col;

	point_line(line);
	if (current_buff != -1)
		temp = local_table[current_buff].segment;
	else
		temp = -1;
	if (temp != -1)
		file_p = global_table[temp].file_offset;
	else
		return 0;

	lseek(in_file, file_p, SEEK_SET);
	read(in_file, disk_buffer, DISKBUFFER);
        line -= global_table[temp].first_line;

	p = disk_buffer;
	ch = -1;
	while (line--)
	{
		col = 0;
		for (;;)
		{
			if (ch == -1)
			{
				ch = *p++;
			}
			switch (ch)
			{
			case '\r':
				ch = -1;
				file_p++;
				break;
			case '\t':
				col++;
				while(col % 8 && col < COLUMNS)
					col++;
				ch = -1;
				file_p++;
				if (col == COLUMNS)
					goto new_line;
				break;
			case '\n':
			case FF:
				ch = -1;
				file_p++;
				goto new_line;
			default:
				if (col >= COLUMNS)
					goto new_line;
                        	col++;
				ch = -1;
				file_p++;
				break;
			}
		}
	new_line:  ;
	}
	return file_p;
}

/* --------------------------------------------------------------------------
Display one line at screen line y.  If line does not exist, clear display
line.*/
void display_line(int y, long n)
{
	if (n > last_line)
	{
		clear_line(y);
		return;
	}
        write_line(y, (char *)point_line(n));
}

/* --------------------------------------------------------------------------
Compare line ln with compiled regular expression. If exp occurs in line,
return non-zero, else 0.
*/
int compare(long ln)
{
	int i;
	char *p;
	char *braslist[9], *braelist[9];
        char bodge[81];

	p = point_line(ln);
	memcpy(bodge, p, COLUMNS);
	bodge[80] = 0;
	for (i=79; i>0; i--)
	{
		if (bodge[i] != ' ')
			break;
		bodge[i] = 0;
	}

	return execute(0, bodge, braslist, braelist, ignore_case);
}

/* --------------------------------------------------------------------------
Search forward from line n to find key.  Return -1 if not found, else line
where found.
*/
long search_forward(long line, char *key)
{
	long ln;
	char *p;

	if (key[0] == 0)
		return line > 0 ? line-1 : line ;

	dprintf(1, statusline, "searching \\/ for >%s<", key);
	p = key;
	compile(&p);

	for (ln = line; ln <= last_line; ln++)
	{
		if (compare(ln))
			return ln;
	}
	return -1;
}

/* --------------------------------------------------------------------------
Search backward from line n to find key.  Return -1 if not found, else line
where found.
*/
long search_backward(long line, char *key)
{
	long ln;
	char *p;

	if (key[0] == 0)
		return line > 0 ? line-1 : line ;

	dprintf(1, statusline, "searching /\\ for >%s<", key);
	p = key;
	compile(&p);

	for (ln = line; ln >= 0; ln--)
	{
		if (compare(ln))
			return ln;
	}
	return -1;
}

/* --------------------------------------------------------------------------
Initialise data structures for a new file,
*/
void new_file(void)
{
	int i;

	serial = 0;
	for (i=0; i<GLOBALTABLESIZE; i++)
	{
		global_table[i].first_line = 0;
		global_table[i].file_offset = 0;
		global_table[i].in_buffer = -1;
	}
	for (i=0; i<buffers && i<MAXBUFFERS; i++)
	{
		local_table[i].access_serial = 0;
		local_table[i].segment = -1;
	}
	for (i=0; i<BOOKMARKS; i++)
	{
		mark[i] = -1;
	}
	last_line = BIGGEST;
	current_buff = -1;
}

/* --------------------------------------------------------------------------
Display screen starting at line n.
*/
void display_screen(long n)
{
	int i;

	for (i=0; i<uselines; i++)
		display_line(i, n+i);
	return;
}

/* --------------------------------------------------------------------------
Save old text screen.
*/
void save_old_screen(void)
{
	unsigned char *p, *q;
	unsigned char far *d;

	q = p = save_screen; d = disp_mem;
	while (q < p + 2 * rows * COLUMNS)
		*q++ = *d++;
}

/* --------------------------------------------------------------------------
Restore old text screen from pointer
*/
void restore_old_screen(void)
{
	unsigned char *p, *q;
	unsigned char far *d;

	q = p = save_screen; d = disp_mem;
	while (q < p + 2 * rows * COLUMNS)
		*d++ = *q++;
}

/* --------------------------------------------------------------------------
Generate file ID string for index file
*/
char *file_id(int handle, char *fname)
{
	static char id[MAXFILE + MAXEXT + 100];
	char drive[MAXDRIVE], dir[MAXDIR], name[MAXFILE], ext[MAXEXT];
	struct ftime ft;
	long fl;

	fnsplit(fname, drive, dir, name, ext);
	getftime(handle, &ft);
	fl = filelength(handle);
	sprintf(id, "%4.4d-%2.2d-%2.2d = %2.2d:%2.2d:%2.2d = %10ld = %s%s",
		ft.ft_year + 1980, ft.ft_month, ft.ft_day,
		ft.ft_hour, ft.ft_min, ft.ft_tsec * 2,
		fl,
		name, ext);
	
	return id;
}

/* --------------------------------------------------------------------------
Generate filename for index file
*/
char *index_name(char *fname)
{
	static char path[MAXPATH];
	char drive[MAXDRIVE], dir[MAXDIR], name[MAXFILE], ext[MAXEXT];

	fnsplit(fname, drive, dir, name, ext);
	fnmerge(path, drive, dir, name ,".ind");
	return path;
}

/* --------------------------------------------------------------------------
Save status of global_table and bookmarks
*/
void save_status(int handle, char *name)
{
	FILE *indfile;
        int i;

	indfile = fopen(index_name(name), "w");
	if (indfile == NULL)
	{
		dprintf(1, statusline, "cannot open index file");
		get_keyb();
		return;
	}
	fprintf(indfile, TITLE);
	fprintf(indfile, "\nfile_id\n");
	fprintf(indfile, "%s\n", file_id(handle, name));
	fprintf(indfile, "global_table\n");
	for (i=0; i<GLOBALTABLESIZE; i++)
	{
		if (global_table[i].first_line == 0 && i != 0)
			break;
		fprintf(indfile, "%d %ld %ld\n", i,
			global_table[i].first_line,
			global_table[i].file_offset);
	}
	fprintf(indfile, "marks\n");
	for (i=0; i<BOOKMARKS; i++)
		fprintf(indfile, "%d %ld\n", i, mark[i]);
	fprintf(indfile, "last_line\n");
	fprintf(indfile, "%ld\n", last_line);
	fprintf(indfile, "end\n");
	fclose(indfile);
}

/* --------------------------------------------------------------------------
Restore status of global_table and bookmarks.  Assumes that the tables
are as they are set by new_file(), i.e. completely reset.
*/
void restore_status(int handle, char *name)
{
	FILE *indfile;
	char s[200];
        long x, y, z;

        indfile = fopen(index_name(name), "r");
	if (indfile == NULL)
		return;
	fgets(s, 199, indfile);
	if (strncmp(s, TITLE, strlen(TITLE)) != 0)
		goto leave;
	fgets(s, 80, indfile);
	if (strcmp(s, "file_id\n") != 0)
		goto leave;
	fgets(s, 80, indfile);
	if (strncmp(s, file_id(handle, name), 
				strlen(file_id(handle, name))) != 0)
		goto leave;
	fgets(s, 80, indfile);
	if (strcmp(s, "global_table\n") != 0)
		goto leave;
	while (fgets(s, 80, indfile) != NULL && strcmp(s, "marks\n") != 0)
	{
		sscanf(s, "%ld %ld %ld", &x, &y, &z);
		global_table[x].first_line = y;
		global_table[x].file_offset = z;
	}
        if (strcmp(s, "marks\n") != 0)
	{
		new_file();
		goto leave;
	}
	while (fgets(s, 80, indfile) != NULL && strcmp(s, "last_line\n") != 0)
	{
		sscanf(s, "%ld %ld", &x, &y);
		mark[x & 0x1f] = y;
	}
	if (strcmp(s, "last_line\n") != 0)
	{
		new_file();
		goto leave;
	}
	fgets(s, 80, indfile);
	sscanf(s, "%ld", &last_line);
	fgets(s, 80, indfile);
	if (strcmp(s, "end\n") != 0)
	{
		new_file();
		goto leave;
	}
leave:
	fclose(indfile);
}
/* --------------------------------------------------------------------------
Help for text mode.
*/
void help2(void)
{
	clear_screen();
	dprintf(1, 0, TITLE);
	dprintf(1, 1,  "  commands:");
	dprintf(1, 2,  "CR CurDn e j  one more line");
	dprintf(1, 3,  "SP PagDn f    one more screen");
       	dprintf(1, 4,  "<n>G <n>g     goto a line");
	dprintf(1, 5,  "h F1          this help");
	dprintf(1, 6,  "q             quit this file, read next if any");
	dprintf(1, 7,  "ESC           quit all files");
	dprintf(1, 8,  "g t HOME      top");
	dprintf(1, 9, "G END         bottom");
	dprintf(1, 10, "CurUp y k     back a line");
	dprintf(1, 11, "PagUp b       back a screen");
	dprintf(1, 12, "d u           down, up <n> lines, default 10");
	dprintf(1, 13, "m<a> '<a>     set, go-to a mark");
	dprintf(1, 14, "<n>p <n>%%     go-to percent into file");
	dprintf(1, 15, "<n>a          change colour");
	dprintf(1, 16, "s r           save index, repair buffers");
	dprintf(1, 17, "/<text>\\r n   search forward, repeat");
	dprintf(1, 18, "?<text>\\r N   search backward, repeat");
	dprintf(1, 19, "x             switch to hex mode");
	dprintf(1, 20, "<n>i          odd <n> makes case insensitive");
	dprintf(1, 21, "most move commands will take prefix <n>");
  	dprintf(1, 22, "(using %d buffers)", buffers);
	dprintf(1, statusline, "CR to continue");
	get_keyb();
}

/* --------------------------------------------------------------------------
Get search text to search key buffer.
*/
void get_key(void)
{
	char *p, ch;

	p = search_key;
	*p = '\0';
	for (;;)
	{
		dprintf(1, statusline, "search for: >%s<", search_key);
		ch = get_keyb() & 0xff;
		if (ch == '\r' || p >= search_key + COLUMNS)
			break;
		else if (ch == BS && p > search_key)
			*(--p) = '\0';
		else
		{
			*p++ = ch;
			*p = '\0';
		}
	}
}

/* --------------------------------------------------------------------------
Main loop, displaying text and getting input. Returns non-zero if escape
or control-c entered, zero otherwise. (Escape and control-C exit out of
a batch of files, q only from the one we're in.) Also returns non-zero if
a file error occurs.
*/
int filter_work(char *filename)
{
	unsigned tc;
	int end;
	long f_size, disp_line, number, temp;

	save_mode = get_video_mode();
	if (save_mode != WANTED_MODE)
		set_video_mode(WANTED_MODE);
        else
		save_old_screen();

	if (bios == 0 && attrib == -1)
		attrib = *(disp_mem + 1);
	clear_screen();
	dprintf(1, 1, TITLE);
	dprintf(1, 2, COPR);

	if (strcmp(filename, "stdin") == 0)
		in_file = 0;
	else
		in_file = _open(filename, O_RDONLY);
	if (in_file < 0)
	{
		dprintf(1, statusline, "can't open %s", filename);
		get_keyb();
		goto bibi;
	}
	if (isatty(in_file))
	{
		dprintf(1, statusline,
			"cannot read from a terminal, use '-h' for help");
		get_keyb();
		goto bibi;
	}

	f_size = filelength(in_file);

	setjmp(jump_buff);

       	new_file();
	end = 0;
	disp_line = 0;
	number = 0;
	restore_status(in_file, filename);

	while (!end)
	{
		if (number == 0)
			display_screen(disp_line);

		if (last_line == BIGGEST || last_line == 0)
			dprintf(1, statusline,
				"z - %s line %ld (%ld bytes) (ic%d) [%ld]",
				filename, disp_line+1, f_size,
				ignore_case, number);
		else
                 	dprintf(1, statusline,
				"z - %s line %ld of %ld (%.0f%%) (ic%d) [%ld]",
				filename, disp_line+1, last_line+1,
				(double)disp_line*100./last_line,
				ignore_case, number);

		tc = get_keyb();
		if (tc & 0xff)
			tc &= 0xff;

		switch (tc)
		{
		case '0': case '1': case '2': case '3': case '4':
		case '5': case '6': case '7': case '8': case '9':
			number = number * 10 + (tc - '0');
			if (number > 200000000L)
				number = 200000000L;
			break;
		case ' ':
		case PAGEDOWN:
		case 'f':
			if (number == 0)
				number = 1;
			disp_line += uselines * number;
			if (disp_line > last_line)
				disp_line = last_line;
			number = 0;
			break;
		case 'd':
			if (number != 0)
				udmove = number;
			disp_line += udmove;
			if (disp_line > last_line)
				disp_line = last_line;
			number = 0;
			break;
		case 'u':
			if (number != 0)
				udmove = number;
			disp_line -= udmove;
			if (disp_line < 0)
				disp_line = 0;
			number = 0;
			break;
		case 't':
		case HOME:
			disp_line = 0;
			number = 0;
			break;
		case 'b':
		case PAGEUP:
			if (number == 0)
				number = 1;
			disp_line -= uselines * number;
			if (disp_line < 0)
				disp_line = 0;
			number = 0;
			break;
		case '\r':
		case CURRDOWN:
		case 'e':
		case 'j':
			if (number == 0)
				number = 1;
			disp_line += number;
			if (disp_line > last_line)
				disp_line = last_line;
			number = 0;
			break;
		case END:
			number = 0;
			goto big_g;
		case 'g':
			if (number == 0)
				number = 1;
			goto big_g;
		case 'G':  big_g:
			number--;
			if (number < 0 && last_line == BIGGEST)
				display_line(0, BIGGEST);
			if (number < 0)
				number = last_line - uselines + 1;
			if (number < 0)
				number = 0;
			if (number > last_line)
				number = last_line;
			disp_line = number;
			number = 0;
			break;
		case CURRUP:
		case 'y':
		case 'k':
			if (number == 0)
				number = 1;
			disp_line -= number;
			if (disp_line < 0)
				disp_line = 0;
			number = 0;
			break;
		case 'h':
		case F1:
			help2();
			number = 0;
			break;
		case 'i':
			ignore_case = 1 & number;
			number = 0;
			break;
		case 'a':
			if (number != 0)
			{
				attrib = number & 0xff;
                       	}
			else
			{
				if ((attrib & 0x07) != 0x07)
					attrib++;
				else if ((attrib & 0x70) != 0x70)
					attrib = (attrib + 0x10) &
							0xf8;
				else
					attrib = 0x00;  /* !! */
			}
			clear_screen();
			number = 0;
			break;
		case 'q':
			end = 1;
			break;
		case ESCAPE:
			end = 2;
			break;
		case '/':
		case 'S'-0x40:
			get_key();
		case 'n':
			if (number == 0)
				number = 1;
			if (disp_line >= last_line)
				number = 0;
			while (number--)
			{
				temp = search_forward(disp_line + 1,
					search_key);
				if (temp != -1)
					disp_line = temp;
				else
					break;
			}
			number = 0;
			break;
		case '?':
		case 'R'-0x40:
			get_key();
		case 'N':
		case ',':
			if (number == 0)
				number = 1;
             			if (disp_line <= 0)
				number = 0;
			while (number--)
			{
				temp = search_backward(disp_line - 1,
						search_key);
				if (temp != -1)
					disp_line = temp;
				else
					break;
			}
			number = 0;
			break;
		case 'm':
			dprintf(1, statusline, "set mark > <");
			mark[get_keyb() & 0x1f] = disp_line;
			number = 0;
			break;
		case '\'':
			dprintf(1, statusline, "goto mark > <");
			temp = get_keyb() & 0x1f;
			if (mark[temp] != -1)
				disp_line = mark[temp];
			number = 0;
			break;
		case 'p': case '%' :
			if (number > 100)
				number = 100;
			if (last_line == BIGGEST)
				display_line(0, BIGGEST);
			disp_line = last_line * number / 100;
			number = 0;
			break;
		case 's':
			save_status(in_file, filename);
			number = 0;
			break;
		case 'r':
			new_file();
			number = 0;
			break;
		case 'x':
			hex_mode(in_file, line_in_file(disp_line), filename);
			break;
		default:
			number = 0;
			break;
		}
	}
bibi:
	if (in_file != 0)
		_close(in_file);
	if (save_mode != WANTED_MODE)
		set_video_mode(save_mode);
	else
		restore_old_screen();
	return end - 1;
}

/* --------------------------------------------------------------------------
Help, if requested on the command line by - sequence.
*/
void help1(char *name)
{
	printf(TITLE); printf("\n");
	printf(COPR);
	printf("\n\nUsage:\n");
	printf("  %s <opts> <path>... to read from files\n", name);
	printf("                         (wildcards allowed)\n");
	printf("  %s <opts>           to read stdin\n", name);
	printf("                         (or filename == '-' or 'stdin')\n");
	printf("  -h                  to get this help\n");
	printf("  -a<n>               to use different display attribute\n");
	printf("  -b                  to force BIOS calls\n");
	printf("  -i<n>               n=0 for case sensitive (def insens)\n");
	printf("  -m                  use memory at b0000 not b8000\n");
	printf("\nAt the command prompt: F1 for help\n\n");
}

/* --------------------------------------------------------------------------
Main, command line argument decoding only.
*/
int main(int argc, char **argv)
{
	int i, files_found;
	char *s;

	argv = wildexpand(&argc, argv);
	files_found = 0;
	attrib = -1;
	bios = 0;

	disp_mem = NORM_MEM;
	udmove = DEFAULTUDMOVE;
	ignore_case = 1;

	s = getenv("ROWS");
	if (s == NULL || (rows = atoi(s)) < 10 || rows > 100)
		rows = DEFAULTROWS;
	statusline = rows - 1;
	uselines = rows - 1;

	global_table = malloc(GLOBALTABLESIZE *
				sizeof(struct global_table_entry));
        if (global_table == NULL)
	{
		fprintf(stderr, "Cannot malloc for global table\n");
		return 2;
	}
	local_table = malloc(MAXBUFFERS *
				sizeof(struct local_table_entry));
	if (local_table == NULL)
	{
		fprintf(stderr, "Cannot malloc for local table\n");
		return 2;
	}
	disk_buffer = malloc(DISKBUFFER);
	if (disk_buffer == NULL)
	{
		fprintf(stderr, "Cannot malloc for disk_buffer\n");
		return 2;
	}
	save_screen = malloc(2 * rows * COLUMNS);
	if (save_screen == NULL)
	{
		fprintf(stderr, "Cannot malloc for save_screen\n");
		return 2;
	}
	buffers = 0;
	while (buffers < MAXBUFFERS)
	{
		local_table[buffers].buffer_location = malloc(BUFFERSIZE);
		if (local_table[buffers].buffer_location == NULL)
			break;
		buffers++;
	}
	if (buffers > 0)
	{
		free(local_table[--buffers].buffer_location);
	}
	if (buffers == 0)
	{
		fprintf(stderr, "Cannot get even one buffer\n");
		return 2;
	}

	for (i=1; i<argc; i++)
	{
		if (argv[i][0] != '-')
		{
			files_found++;
			if (filter_work(argv[i]))
				return 0;
		}
		else
		{
			switch(argv[i][1])
			{
			case '?':  case 'h':
				help1(argv[0]);
				return 2;
			case 'b':
				bios = 1;
				break;
			case 'a':
				attrib = atoi(argv[i]+2);
				break;
			case 'i':
				ignore_case = 1 & atoi(argv[i]+2);
				break;
			case 'm':
				disp_mem = HERC_MEM;
				break;
			case '\0':
				files_found++;
				filter_work("stdin");
				break;
			default:
				fprintf(stderr, "invalid switch\n");
				help1(argv[0]);
				return 2;
			}
		}
	}
	if (files_found == 0)
		filter_work("stdin");
	return 0;
}
